/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/**
 * ASTEncoder.java
 * Created:
 *
 * @author  Valter Crescenzi
 * @version
 */

package roadrunner.ast;

import java.io.Writer;
import java.io.PrintWriter;
import java.util.*;
import java.util.logging.*;

import roadrunner.util.Indenter;
import roadrunner.parser.Token;
import roadrunner.config.Config;
import roadrunner.config.Preferences;
import roadrunner.config.Constants;


public class ASTEncoder extends VisitorPlugger implements ASTConstants {
    
    static private Logger log = Logger.getLogger(ASTEncoder.class.getName());
    
    protected PrintWriter out;
    protected Indenter ind;
    
    public ASTEncoder(Indenter ind, Writer out) {
        this.ind = ind;
        this.out = new PrintWriter(out);
    }
    
    public void encode(Expression exp) {
        out.println(ind+"<"+EXPRESSION+">");
        encode(exp.getRoot());
        out.println(ind+"</"+EXPRESSION+">");
    }
    
    protected void encode(ASTAnd root) {
        this.visit(root);
    }
    
    public boolean visit(ASTAnd node) {
        return serializeStructureNode(node);
    }
    public boolean visit(ASTPlus node) {
        return serializeStructureNode(node);
    }
    public boolean visit(ASTHook node) {
        return serializeStructureNode(node);
    }
    
    private boolean serializeStructureNode(SimpleNode node) {
        ind.inc();
        out.println(ind+"<"+node.getType()+">");
        node.childrenAccept(this);
        out.println(ind+"</"+node.getType()+">");
        ind.dec();
        return false;
    }
    
    public boolean visit(ASTSubtree node) {
        ind.inc();
        out.print(ind+"<"+node.getType()+">");
        encodeToken(node.getRootToken());
        out.println("</"+node.getType()+">");
        ind.dec();
        return false;
    }
    
    public boolean visit(ASTVariant node) {
        ind.inc();
        out.print(ind+"<"+VARIANT+" "+LABEL+"="+"\""+node.getLabel()+"\">");
        encodeToken(node.getToken());
        out.println("</"+VARIANT+">");
        ind.dec();
        return false;
    }
    
    public boolean visit(ASTToken token) {
        ind.inc();
        out.print(ind);
        encodeToken(token);
        out.println();
        ind.dec();
        return false;
    }
    
    private void encodeToken(Token token) {
        if (token.isTag()) encodeTag(token);
        else encodeText(token);
    }
    
    private void encodeTag(Token tag) {
        out.print("<"+TAG+" "+ELEMENT+"=\""+ (tag.isEndTag()?"/":"")+tag.getElement()+"\"");
        out.print(" "+DEPTH+"=\""+tag.depth()+"\"");
        encodeAttributes(tag);
        out.print("/>");
    }
    
    private void encodeAttributes(Token tag) {
        Map attrs = tag.getAttributes();
        out.print(" "+ATTRIBUTES+"=\"");
        Iterator attrsIt = attrs.entrySet().iterator();
        if (attrsIt.hasNext()) encodeAttribute((Map.Entry)attrsIt.next());
        while (attrsIt.hasNext()) {
            out.print(ATTRIBUTES_SEPARATOR);
            encodeAttribute((Map.Entry)attrsIt.next());
        }
        out.print("\"");
    }
    
    private void encodeAttribute(Map.Entry entry) {
        out.print(entry.getKey().toString());
        if (entry.getValue()!=null)
            out.print(ATTRIBUTE_NAME_VALUE_SEP+entry.getValue().toString());
    }
    
    private void encodeText(Token text) {
        out.print("<"+PCDATA+" "+DEPTH+"=\""+text.depth()+"\">");
        if (text.getText()!=null) {
            //We need CDATA sections because free-text could contain tags and other stuff
            out.print("<![CDATA["+text.getText()+"]]>");
        }
        out.print("</"+PCDATA+">");
    }
    
    
}
