/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * Region.java
 *
 * Created on 25 gennaio 2003, 12.01
 * @author  Valter Crescenzi
 */
package roadrunner.ast;

import java.util.List;

import roadrunner.bidi.Region;

public class ExpressionRegion extends Region {
    
    private Expression exp;
    private Node.Path base; // The ASTAnd node which is the base of the region
    
    /** Creates a new instance of Region for AST */
    public ExpressionRegion(Expression exp, Node.Path base, int left, int right) {
        super(left,right);
        this.exp  = exp;
        this.base = base;
        checkRegion(exp,base,left,right);
    }
    
    /** Creates a new instance of Region for all children of an ASTAnd node */
    public ExpressionRegion(Expression exp, Node.Path base) {
        this(exp, base, 0, exp.getNode(base).jjtGetNumChildren());
    }
    
    public Expression getExpression() { return this.exp; }
    
    public boolean isFull() { return this.size() == getBase().jjtGetNumChildren(); }
    public ASTAnd getBase() { return getBase(getExpression());  }
    
    public ASTAnd getBase(Expression exp) {
        return (ASTAnd)exp.getNode(this.base);
    }
    
    public ExpressionRegion getBaseRegion() {
        return new ExpressionRegion(getExpression(),getBasePath());
    }
    
    public Node.Path getBasePath() {
        return this.base;
    }
    
    public Node.Path getAbsolutePath(Node.Path relative) {
        if (relative.depth()==0) return getBasePath();
        int relativeIndices[] = relative.indices();
        relativeIndices[0]+= this.getLeftBorderPos(); // pos -> index
        return getBasePath().absolute(new PathArray(relativeIndices));
    }
    
    public Region subRegion(int l, int r) {
        return new ExpressionRegion(getExpression(), getBasePath(), l, r);
    }
    
    public List asList() { return this.asList(this.exp); }
    
    public List asList(Expression exp) {
        // Cosi' si possono fare modifiche tramite regioni
        return this.getBase(exp).jjtGetChildren().subList(getLeftBorderPos(),getRightBorderPos());
    }
    
    public boolean isBorderRegion() {
        return getLeftBorderPos()==0 || getRightBorderPos()==getBase().jjtGetNumChildren();
    }
    
    private static void checkRegion(Expression exp, Node.Path basePath, int left, int right) {
        Node base = exp.getNode(basePath);
        if (!(base instanceof ASTAnd))
            throw new IllegalArgumentException("Base of region must be an ASTAnd node instead of "+base);
        int children = base.jjtGetNumChildren();
        checkBorders(children, left,right);
    }
      
    public boolean equals(Object o) {
        ExpressionRegion r = (ExpressionRegion)o;

        return this.getLeftBorderPos()==r.getLeftBorderPos() && this.getRightBorderPos()==r.getRightBorderPos() &&
               this.getBasePath().equals(r.getBasePath()) && this.exp.equals(r.exp);
    }
    
    public int hashCode() {
        return this.exp.hashCode() + this.getBasePath().hashCode() + this.getLeftBorderPos() + this.getRightBorderPos();
    }
    
}
