/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * PathArray.java
 *
 * Created on 
 * @author  Valter Crescenzi
 */
package roadrunner.ast;

import java.util.*;

public class PathArray implements Node.Path, Comparable {
    
    private int[] indices;
    private int depth;
    
    public PathArray(int[] indices) {
        this.indices = new int[indices.length];
        System.arraycopy(indices, 0, this.indices, 0, indices.length);
        this.depth = indices.length;
    }
    
    public int[] indices() {
        int result[] = new int[this.depth];
        System.arraycopy(this.indices, 0, result, 0, this.depth);
        return result;
    }
    
    public int lastIndex() {
        if (this.depth<1) throw new IllegalArgumentException("This path is empty!");
        return this.indices[this.depth-1];
    }
    
    public int depth() {
        return this.depth;
    }
    
    public Node.Path subpath(int depth) {
        //    depth < 0 means to remove -depth indices from path        
        int d = depth;
        if (Math.abs(depth)>this.depth)
            throw new IllegalArgumentException("A path of depth "+this.depth+" does not contains a subpath of depth "+depth);
        if (depth<0) d = this.depth+depth;
        int[] subindices = new int[d];
        System.arraycopy(this.indices, 0, subindices, 0, d);
        return new PathArray(subindices);
    }
    
    public Node.Path absolute(Node.Path relativePath) {
        // Make the absolute node path of the node reacheable
        // starting from this path by applying a relative path
        int thisIndices[] = this.indices();
        int relativeIndices[] = relativePath.indices();
        int absoluteIndices[] = new int[thisIndices.length+relativeIndices.length];
        System.arraycopy(thisIndices, 0, absoluteIndices, 0, thisIndices.length);
        System.arraycopy(relativeIndices, 0, absoluteIndices, thisIndices.length, relativeIndices.length);
        return new PathArray(absoluteIndices);
    }
    
    public boolean equals(Object o) {
        Node.Path p = (Node.Path)o;
        return Arrays.equals(this.indices(),p.indices());
    }
    
    public int hashCode() {
        int result = 0;
        for(int i=0; i<this.depth; i++)
            result+=this.indices[i];
        return result;
    }
    
    
    public int compareTo(Object o) {
        Node.Path p = (Node.Path)o;
        int min = Math.min(this.depth(), p.depth());
        int[]  i1 = this.indices();
        int[]  i2 = p.indices();
        for(int k=0; k<min; k++) {
            if (i1[k]>i2[k]) return +1;
            else if (i1[k]<i2[k]) return -1;
        }
        if (this.depth()>p.depth()) return +1;
        else if (this.depth()<p.depth()) return -1;
        return 0;
    }
    
    public String toString() {
        StringBuffer result = new StringBuffer();
        for(int i=0; i<this.depth; i++) {
            result.append("."+this.indices[i]);
        }
        return result.toString();
    }
    
    
}
