/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * PrefixHunter.java
 *
 * Created on 4 ottobre 2003, 10.23
 * @author  Valter Crescenzi
 */

package roadrunner.ast;

import java.util.*;
import java.util.logging.*;

import roadrunner.bidi.*;
import roadrunner.parser.Token;

public class PrefixHunter extends VisitorPlugger implements Visitor {
    
    static private Logger log = Logger.getLogger(PrefixHunter.class.getName());
    
    static public Node getFirstLeaf(Direction dir, Node node) {
        while (node.jjtGetNumChildren()>0) {
            List kids = node.jjtGetChildren();
            node = (Node)BidirectionalListFactory.newListView(dir, kids).getFirst();
        }
        return node;
    }
    
    /** Creates a new instance of PrefixHunter */
    private int k;
    private List currentPrefixes;
    private List finalPrefixes;
    
    public PrefixHunter(int k, Direction dir) {
        super(dir);
        this.k = k;
        this.currentPrefixes = new LinkedList();
    }
    
    public int getK() {
        return this.k;
    }
    
    public List getPrefixes(Region region) {
        this.finalPrefixes = new LinkedList();
        this.currentPrefixes.add(new ArrayList());
        visit(region);
        this.finalPrefixes.addAll(this.currentPrefixes); // these are shorter than k tokens
        this.currentPrefixes.clear();
        return this.finalPrefixes;
    }
    
    /* methods return true if prefixes are not completed (length < k) */
    public boolean visit(ASTAnd node) {
        boolean finished = false;
        ListIterator it = BidirectionalListFactory.newListIterator(node.jjtGetChildren(), this.dir);
        while (it.hasNext() && !finished) {
            Node child = (Node)it.next();
            finished = child.jjtAccept(this);
        }
        return finished;
    }
    public boolean visit(ASTPlus node) {
        boolean finished = visitNode(node);
        if (!finished) {
            List wasCurrentPrefixes = getCurrentPrefixesClone();
            visitNode(node);
            this.currentPrefixes.addAll(wasCurrentPrefixes); // prefixes for case in which plus node doesn't' match once more
            return false; //not finished: there are these ^ prefixes to complete
        }
        return true;
    }
    public boolean visit(ASTHook node) {
        List wasCurrentPrefixes = getCurrentPrefixesClone();
        visitNode(node);
        this.currentPrefixes.addAll(wasCurrentPrefixes); // prefixes for case in which hook node doesn't match
        return false; //not finished: there are these ^ prefixes to complete
    }
    public boolean visit(ASTSubtree node) {
        return addTokenToPrefixes(null);
    }
    public boolean visit(ASTVariant node) {
        return addTokenToPrefixes(node.getToken());
    }
    public boolean visit(ASTToken node) {
        return addTokenToPrefixes(node);
    }
    
    private List getCurrentPrefixesClone() {
        List result = new LinkedList();
        Iterator it = this.currentPrefixes.iterator();
        while (it.hasNext()) {
            List prefix = (List)it.next();
            result.add(new ArrayList(prefix));
        }
        return result;
    }
    
    private boolean addTokenToPrefixes(Token token) {
        boolean finished = true;
        Iterator it = this.currentPrefixes.iterator();
        while (it.hasNext()) {
            List prefix = (List)it.next();;
            prefix.add(token);
            if (prefix.size()==k) {
                this.finalPrefixes.add(prefix);
                it.remove();
            }
            else finished = false;
        }
        return finished;
    }
}
