/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * BidirectionalList.java
 * This class provides both left to right and right to left view
 * over an underlying java.util.List. This way, it overcomes the
 * limit of the standard interface which impose to the client programmers
 * a left to right view.
 *
 * Created on 30 gennaio 2003, 12.28
 * @author  Valter Crescenzi
 */

package roadrunner.bidi;

import java.util.*;

/**
 */
public class BidirectionalListFactory {
    
    public static BidirectionalList newListView(Direction dir, final List l) {
        return (dir==Direction.RIGHT2LEFT ?
        (BidirectionalList)new ReverseList(l) :
            (BidirectionalList)new DirectList(l)  );
    }
    
    public static ListIterator newListIterator(List list, Direction dir) {
        BidirectionalList bidilist =  BidirectionalListFactory.newListView(dir,list);
        return bidilist.listIterator();
    }
    
    static private abstract class ListDelegator implements List {
        protected List list;
        protected ListDelegator(final List l) {
            if (l instanceof BidirectionalList)
                throw new RuntimeException("Cannot build a bidirectional view over a bidirectional view!");
            this.list = l;
        }
        public boolean add(Object o)               { return list.add(o); }
        public void add(int index, Object element) { list.add(index,element); }
        public boolean addAll(Collection c)        { return list.addAll(c); }
        public boolean addAll(int index, Collection c) { return list.addAll(index,c); }
        public void clear()                        { list.clear();}
        public boolean contains(Object o)          { return list.contains(o); }
        public boolean containsAll(Collection c)   { return list.containsAll(c); }
        public Object get(int index)              { return list.get(index); }
        public int indexOf(Object o)               { return list.indexOf(o); }
        public boolean isEmpty()           { return list.isEmpty(); }
        public Iterator iterator()         { return listIterator(); }
        public int lastIndexOf(Object o)   { return list.lastIndexOf(o); }
        public ListIterator listIterator() { return list.listIterator(); }
        public ListIterator listIterator(int index) { return list.listIterator(index); }
        public boolean remove(Object o) { return list.remove(o); }
        public Object remove(int index) { return list.remove(index); }
        public boolean removeAll(Collection c) { return list.removeAll(c); }
        public boolean retainAll(Collection c) { return list.retainAll(c); }
        public Object set(int index, Object element) { return list.set(index,element); }
        public int size()                            { return list.size();}
        public List subList(int fromPos, int toPos)  { return list.subList(fromPos,toPos); }
        public Object[] toArray()           { return list.toArray(); }
        public Object[] toArray(Object[] a) { return list.toArray(a); }
    }
    
    static private class DirectList extends ListDelegator implements BidirectionalList {
        private DirectList(List l) {
            super(l);
        }
        public ListIterator listIterator(Direction dir) {
            if (dir==Direction.LEFT2RIGHT) return listIterator();
            return listIterator(dir,list.size());
        }
        public ListIterator listIterator(Direction dir, int pos) {
            if (dir==Direction.LEFT2RIGHT) return listIterator(pos);
            ListIterator lit = list.listIterator(dir.indexBeforePos(pos));
            ReverseIterator rit = new ReverseIterator(lit);
            return rit;
        }
        public List subListFrom(int position) {
            return list.subList(position,list.size());
        }
        public List subListTo(int position) {
            return list.subList(0, position);
        }
        public List subListToIndex(int index) {
            return list.subList(0, index);
        }
        public Object getFirst() {
            return list.get(0);
        }
        public Object getLast() {
            return list.get(list.size()-1);
        }
    }
    
    static private class ReverseList extends ListDelegator implements BidirectionalList {
        private ReverseList(List l) {
            super(l);
        }
        public boolean add(Object o) {
            this.list.add(0,o);
            return true;
        }
        
        public void add(int pos, Object element)       { throw new UnsupportedOperationException(); }
        public boolean addAll(Collection c)            { throw new UnsupportedOperationException(); }
        public boolean addAll(int index, Collection c) { throw new UnsupportedOperationException(); }
        public Iterator iterator()         { return listIterator(); }
        public ListIterator listIterator() { return listIterator(Direction.RIGHT2LEFT); }
        
        public ListIterator listIterator(Direction dir) {
            if (dir==Direction.LEFT2RIGHT) return this.list.listIterator();
            return listIterator(dir, this.list.size());
        }
        
        public ListIterator listIterator(Direction dir, int pos) {
            if (dir==Direction.LEFT2RIGHT) return this.list.listIterator(pos);
            ListIterator lit = this.list.listIterator(dir.indexBeforePos(pos));
            ReverseIterator rit = new ReverseIterator(lit);
            return rit;
        }
        
        public ListIterator listIterator(int index) {
            ListIterator lit = this.list.listIterator(index+1);
            ReverseIterator rit = new ReverseIterator(lit);
            return rit;
        }
        
        public List subListFrom(int position) {
            return this.list.subList(0,position);
        }
        public List subListTo(int position) {
            return this.list.subList(position,this.list.size());
        }
        public List subListToIndex(int index) {
            return this.list.subList(index+1,this.list.size());
        }
        public Object getFirst() {
            return list.get(list.size()-1);
        }
        public Object getLast() {
            return list.get(0);
        }
        
    }
    
    static private class ReverseIterator implements ListIterator {
        private ListIterator lit;
        ReverseIterator(ListIterator lit) { this.lit = lit; }
        
        public void add(Object o)    { throw new UnsupportedOperationException(); }
        public boolean hasNext()     { return this.lit.hasPrevious(); }
        public boolean hasPrevious() { return this.lit.hasNext(); }
        public Object next()        { return this.lit.previous(); }
        public int nextIndex()      { return this.lit.previousIndex(); }
        public Object previous()    { return this.lit.next();  }
        public int previousIndex()  { return this.lit.nextIndex(); }
        public void remove()       { this.lit.remove(); }
        public void set(Object o)  { this.lit.set(o); }
    }
}
