/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * Backtracker.java
 * A backtracker based on ExpressionIterator
 * Created on December 29, 2003, 3:23 PM
 * @author  Valter Crescenzi
 */

package roadrunner.engine;

import java.util.*;

import roadrunner.ast.Expression;

public abstract class Backtracker extends ExpressionIterator {
    
    private Stack stack;
    private boolean enabled;
    private ExpressionIterator current;
    
    /** Creates a new instance of Backtracker */
    protected Backtracker(Expression initial) {
        this(iteratorOverSingleExp(initial));
    }
    
    protected Backtracker(ExpressionIterator initial) {
        this.stack = new Stack();
        this.stack.push(initial);
        this.enabled = true;
        this.current = null;
    }
    
    public boolean backtrackingEnabled() {
        return enabled;
    }
    public void setBacktracking(boolean enabled) {
        this.enabled = enabled;
    }
    
    protected abstract ExpressionIterator successor(Expression exp);
    protected abstract boolean goalTest(Expression exp);
    
    protected void start() {}
    protected void back() {}
    protected Expression gotFinal(Expression exp) { return exp;}
    protected void expired() {}
    
    protected Expression computeNext() {
        start();
        while (!stack.isEmpty()) {
            this.current = (ExpressionIterator)stack.peek();
            try {
                if (!current.hasNext()) {
                    if (!backtrackingEnabled()) break;
                    stack.pop();
                    back();
                    continue;
                }
                
                Expression result = current.next();
                if (goalTest(result)) {
                    return gotFinal(result);
                }
                else {
                    ExpressionIterator next = successor(result);
                    this.stack.push(next);
                }
            }
            catch (FailException f) {
                // fail() has been called: ignore current "result"
            }
        }
        expired();
        return null;
    }
    
    protected ExpressionIterator getCurrent() {
        return this.current;
    }
    
    protected void fail() {
        throw new FailException();
    }
    
    protected class FailException extends RuntimeException {
        protected FailException() {
        }
    }
    
}
