/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * Engine.java
 * 
 * Created on 
 * @author  Valter Crescenzi
 */
package roadrunner.engine;

import java.util.*;

import java.util.logging.*;

import roadrunner.Sample;
import roadrunner.Wrapper;
import roadrunner.WrappersRepository;
import roadrunner.bidi.Direction;
import roadrunner.config.*;
import roadrunner.ast.Expression;
import roadrunner.parser.Parser;
import roadrunner.parser.TokenList;
import roadrunner.engine.space.SearchSpace;

public class Engine implements Constants {
    
    static private Logger log = Logger.getLogger(Engine.class.getName());
    
    static final private Preferences step0prefs = new Preferences(); //Preferences for step 0 (first of all, find subtrees)
    static  {
        //Set preferences for step 0
        step0prefs.setName("Step 0 Preferences");
        step0prefs.put(HOOK,   "false");
        step0prefs.put(PLUS,   "false");
        step0prefs.put(SUBTREE, "true");
        step0prefs.put(AMBIGUITY, "0");
        step0prefs.put(SOLUTIONS, "1");
    }
    
    private Expression expression0;
    private TokenList[] tokenlists;
    
    private Sample[] samples;
    
    private int minDOMdepth4subtrees; //useful for recursive step on subtrees
    
    public Engine(Expression expression, TokenList[] tokenlists, int minDOMdepth4subtrees) {
        if (expression==null) {
            log.fine("Creating a canonical wrapper from the first sample");
            this.expression0 = new Expression(tokenlists[0]);
        }
        else this.expression0 = expression;
        this.tokenlists = tokenlists;
        this.samples = null;
        this.minDOMdepth4subtrees = minDOMdepth4subtrees;
    }
    
    public static WrappersRepository infer(Wrapper wrapper, Sample[] samples) {
        TokenList[] tokenlists = asTokenLists(samples);
        Expression expression = (wrapper != null ? wrapper.getExpression() : getInitialExpression(tokenlists));
        log.fine("Initial Expression is\n"+ (expression!=null ? expression.dump("\t"): null));
        Engine engine = new Engine(expression, tokenlists, 0);
        Set solutions = engine.match();
        log.fine("Looking for variants...");
        VariantsHunter variantsHunter = new VariantsHunter(solutions,  tokenlists);
        solutions = variantsHunter.insertVariants();
        log.fine("Detected "+variantsHunter.getAmbiguousExpressions().size()+ " ambiguous wrappers");
        Registry registry = new Registry(solutions, tokenlists);
        return new WrappersRepository(registry, samples);
    }
    
    static private TokenList[] asTokenLists(Sample[] samples) {
        TokenList[] tokenlists = new TokenList[samples.length];
        for(int i=0; i<samples.length; i++) {
            tokenlists[i] = samples[i].getTokenlist();
        }
        return tokenlists;
    }
    
    private static Expression getInitialExpression(TokenList[] tokenlists) {
        if (!Config.getPrefs().getBoolean(STEP0)) return null;
        log.info("Performing step 0");
        Preferences prefs = new Preferences();
        prefs.putAll(Config.getPrefs());  // save old prefs
        Config.load(step0prefs);          // set prefs for step 0
        log.config("Preferences before performing step 0:\n "+Config.getPrefs());
        Engine engine = new Engine(null, tokenlists, 0);
        Set solutions0 = engine.match();  // all-subtrees solution
        Config.load(prefs);               // back to ordinary prefs
        log.config("Preferences after performing step 0:\n "+Config.getPrefs());
        if (solutions0.isEmpty()) return null;
        Expression step0expression = (Expression)solutions0.iterator().next();
        return step0expression;
    }
    
    public Set match()  {
        
        final List matchingTokenlists = Arrays.asList(this.tokenlists);
        log.finer("Matching Expression0:\n" + expression0.dump("\t")+
        "\nNumber of tokenlists to match: " + tokenlists.length);
        SearchSpace space0 = new SearchSpace(expression0, tokenlists[0], minDOMdepth4subtrees);
        space0.setBacktracking(isExtBacktrackingEnabled());
        
        Backtracker runner = new Backtracker(space0) {
            private int sol=0, index = 1;
            
            public ExpressionIterator successor(Expression solution) {
                log.finer("Proceeding with next tokenlist of index "+index);
                SearchSpace space = new SearchSpace(solution, tokenlists[index], minDOMdepth4subtrees);
                space.setBacktracking(isExtBacktrackingEnabled());
                index++;
                return space;
            }
            
            public void back() {
                index--;
                log.finer("Backtracking to tokenlist "+(index-1));
            }
            
            public boolean goalTest(Expression exp) {
                if (! parseAll(exp, matchingTokenlists.subList(0, index)) ) fail();
                return index==matchingTokenlists.size();
            }
            
            public Expression gotFinal(Expression solution) {
                sol++; log.finer("Got solution n."+ sol +" matching with all "+index+" tokenlists");
                return solution;
            }
        };
        
        Set expressions = new HashSet();
        int max_number_of_solutions = maxNumberOfSolutions();
        int sol = 0;
        while (sol++<max_number_of_solutions && runner.hasNext()) {
            expressions.add(runner.next());
            log.finer(expressions.size()+" different solutions are known");
        }
        
        log.finer("Found "+expressions.size()+" solutions");
        log.finer("About to process their subtrees");
        Lumberjack jack = new Lumberjack(expressions, tokenlists);
        return jack.processSubtrees();
    }
    
    private boolean parseAll(Expression exp, List samples) {
        Parser parserlr = new Parser(exp, Direction.LEFT2RIGHT);
        Parser parserrl = new Parser(exp, Direction.RIGHT2LEFT);
        ListIterator it = samples.listIterator();
        while (it.hasNext()) {
            TokenList tokenlist = (TokenList)it.next();
            if (!parserlr.parse(tokenlist)) {
                log.finer("-> Expression does not align with tokenlist "+it.previousIndex());
                log.finer(parserlr.getMismatches().toString());
                return false;
            }
            if (!parserrl.parse(tokenlist)) {
                log.finer("<- Expression does not align with tokenlist "+it.previousIndex());
                log.finer(parserrl.getMismatches().toString());
                return false;
            }
        }
        return true;
    }
    
    private int maxNumberOfSolutions() {
        int max = Config.getPrefs().getInt(SOLUTIONS);
        if (max==0) max = Integer.MAX_VALUE;
        return max;
    }
    
    private boolean isExtBacktrackingEnabled() {
        return Config.getPrefs().getBoolean(Constants.EBACKTRACKING);
    }
    
}