/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/**
 * CharacteristicSampler.java
 *
 *
 * Created: Wed Dec 13 17:49:57 2000
 *
 * @author Valter Crescenzi
 * @version
 */

package roadrunner.engine.sampler;

import java.util.*;
import java.util.logging.*;

import roadrunner.bidi.*;
import roadrunner.ast.*;
import roadrunner.parser.Token;
import roadrunner.parser.token.TokenFactory;

public class CharacteristicSampler extends VisitorPlugger implements Visitor {
        
    static private Logger log = Logger.getLogger(CharacteristicSampler.class.getName());
    
    private Map lowerSampled; // Cardinalities: For Hooks 0; For Plus 1; For Subtree <X/>; For Variant X
    private Map upperSampled; // Cardinalities: For Hooks 1; For Plus 2; For Subtree <Y/>; For Variant X
    
    private CharacteristicSample sample;
    private boolean allComputed;
    private int maxSampleLength;
    private int minDOMdepth;
    private int tokensCounter;
    private Region region;
    private TokenFactory factory;
    
    public CharacteristicSampler(Region region, Direction dir) {
        super(dir);
        this.factory = TokenFactory.getInstance();
        this.region = region;
        this.lowerSampled = new IdentityHashMap();
        this.upperSampled = new IdentityHashMap();
        this.allComputed = false;
        this.sample = null;
        this.maxSampleLength = -1;
        this.minDOMdepth     = -1;
        this.tokensCounter   =  0;
    }
    
    public void setMaxSampleLength(int max) {
        this.maxSampleLength = max;
    }
    
    void setMinDOMdepth(int min) {
        this.minDOMdepth = min;
    }
    
    public CharacteristicSample computeNext() {
        this.sample = null;
        if (!this.allComputed) {
            this.sample = new CharacteristicSample(this.region, this.dir);
            this.tokensCounter = 0;
            visit(this.region);
            this.allComputed = lowerSampled.equals(upperSampled);
        }
        return this.sample;
    }
    
    public boolean visit(ASTPlus node) {
        visitNode(node);
        if (!lowerSampled.containsKey(node)) { //twice ?
            lowerSampled.put(node,null);
        }
        else if (!upperSampled.containsKey(node)) {
            visitNode(node);
            upperSampled.put(node,null);
        }
        return false;
    }
    
    public boolean visit(ASTHook node) {
        if (!upperSampled.containsKey(node)) { //once ?
            visitNode(node);
            upperSampled.put(node,null);
        }
        else if (!lowerSampled.containsKey(node)) {
            lowerSampled.put(node,null);
        }
        else {
            /* Hooks of a Plus require this */
            visitNode(node); //all done: go deeper
        }
        return false;
    }
    
    public boolean visit(ASTSubtree node) {
        Token token = node.getRootToken();
        if (!lowerSampled.containsKey(node)) {
            lowerSampled.put(node,null);
            addToken(createOpenTagToken("X",token));
            addToken(createCloseTagToken("X",token));
        }
        else if (!upperSampled.containsKey(node)) {
            upperSampled.put(node,null);
            addToken(createOpenTagToken("Y",token));
            addToken(createCloseTagToken("Y",token));
        }
        return false;
    }
       
    private Token createOpenTagToken(String el, Token token) {
        return ( dir==Direction.LEFT2RIGHT ? 
                    factory.createOpenTagToken(el, Collections.EMPTY_MAP, token.depth()+1) :
                    factory.createCloseTagToken(el, Collections.EMPTY_MAP, token.depth()+1) );
    }
    private Token createCloseTagToken(String el, Token token) {
        return ( dir==Direction.LEFT2RIGHT ? 
                    factory.createCloseTagToken(el, Collections.EMPTY_MAP, token.depth()+1) :
                    factory.createOpenTagToken(el, Collections.EMPTY_MAP, token.depth()+1) );
    }
    
    public boolean visit(ASTVariant node) {
//        if (!lowerSampled.containsKey(node)) {
//            lowerSampled.put(node,null);//NON VA BENE
//            addToken(node.createSampleToken("X"));
//        }
//        else if (!upperSampled.containsKey(node)) {
//            upperSampled.put(node,null);
//            addToken(node.createSampleToken("Y"));
//        }
//        addToken(node);
        addToken(node.getToken());
        return false;
    }
    
    public boolean visit(ASTToken node) {
        return addToken(node);
    }
        
    protected boolean visitList(List list) {
        // N.B. this is the only method which is invoked on all nodes; visitNode() and visitChildren are not invoked on node of base region
        boolean stop = false;
        ListIterator it = BidirectionalListFactory.newListIterator(list, this.dir);
        while (it.hasNext()) {
            Node node = (Node)it.next();
            if (node.jjtAccept(this)) {
                stop = true;
                break; // Check stop condition
            }
        }
        return stop;
    }
    
    private boolean addToken(Token token) {
        this.tokensCounter++;
        this.sample.addToken(token);
        return (this.tokensCounter==this.maxSampleLength || token.depth()==this.minDOMdepth);
    }
    
} // CharacteristicSampler
