/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * Matcher.java
 * This class encapsulates the recursive nature of the matching technique
 * Created on 25 marzo 2003, 11.00
 * @author  Valter Crescenzi
 */

package roadrunner.engine.space;

import java.util.*;
import java.util.logging.*;

import roadrunner.bidi.Direction;
import roadrunner.ast.Node;
import roadrunner.ast.Expression;
import roadrunner.ast.ExpressionRegion;
import roadrunner.ast.PrefixHunter;

import roadrunner.parser.Token;
import roadrunner.parser.TokenList;

import roadrunner.engine.Backtracker;
import roadrunner.engine.ExpressionIterator;

import roadrunner.engine.sampler.CharacteristicSet;
import roadrunner.engine.sampler.CharacteristicSample;

/**
 *
 * @author  Valter Crescenzi
 */
public class Matcher extends Backtracker {
    
    static private Logger log = Logger.getLogger(Matcher.class.getName());
    
    private Direction  dir;
    
    private ExpressionRegion region;
    private ListIterator samplesIt;
    private ExpressionRegion remainderRegion;
    
    protected CharacteristicSet chi;
    private Set samplesOfRemainder;
    private Set regionsOfRemainder;
    
    private SearchSpace space;
    
    /** Creates a new instance of Matcher */
    public Matcher(ExpressionRegion r, Expression exp, Direction dir, SearchSpace space) {
        this(new CharacteristicSet(r,domDepth4Sampler(dir, exp), dir), exp, dir, space);
    }
    Matcher(CharacteristicSet sampled, Expression exp, Direction dir, SearchSpace space) {
        super(new SearchSpace(exp,sampled.first(), dir, space.getOptions().getMinDOMdepth()));
        this.dir = dir;
        this.space = space;
        this.chi = sampled;
        this.region = (ExpressionRegion)sampled.getRegion();
        this.samplesIt = sampled.listIterator();
        this.samplesIt.next();
    }
    
    private static int domDepth4Sampler(Direction dir, Expression exp) {
        return ((Token)PrefixHunter.getFirstLeaf(dir, (Node)dir.getFirst(exp.asRegion()))).depth()-1;
    }

    protected void start() {
        this.regionsOfRemainder = new LinkedHashSet();
        this.samplesOfRemainder = new LinkedHashSet();
    }
    
    protected void back() {
        this.samplesIt.previous();
    }
    
    protected boolean goalTest(Expression exp) {
        SearchSpace current = (SearchSpace)getCurrent();
        CharacteristicSample aSampleOfRemainder = projectRemainingRegion(current);
        if (aSampleOfRemainder==null) fail();
        this.remainderRegion = (ExpressionRegion)aSampleOfRemainder.getRegion();
        
        regionsOfRemainder.add(aSampleOfRemainder.getRegion());
        if (regionsOfRemainder.size()>1) {
            log.finer("Internal matching is ambiguous ...");
            regionsOfRemainder.remove(aSampleOfRemainder.getRegion());
            fail();
        }
        this.samplesOfRemainder.add(aSampleOfRemainder);
        return !this.samplesIt.hasNext(); // all samples done
    }
    
    protected ExpressionIterator successor(Expression generalized) {
        log.finer("Proceeding with next sample "+this.samplesIt.nextIndex());
        TokenList nextSample = (TokenList)this.samplesIt.next();
        int minDOMdepth = this.space.getOptions().getMinDOMdepth();
        return new SearchSpace(generalized, nextSample, this.dir, minDOMdepth);
    }
    
    protected Expression gotFinal(Expression exp) {
        SearchSpace current = (SearchSpace)getCurrent();
        ExpressionRegion matchingRegion = current.getMatchingRegion();
        if (!this.space.getOperatorFactory().checkBorderNodes(matchingRegion)) fail();
        if (!remainderRegion.getBasePath().equals(region.getBasePath())) fail();
        return new Expression(matchingRegion);
    }
    
    public ExpressionRegion getRemainingRegion() {
        return this.remainderRegion;
    }
    
    public CharacteristicSet getCharacteristicSetOfRemainingRegion() {
        return new CharacteristicSet(this.remainderRegion, this.samplesOfRemainder);
    }
    
    private CharacteristicSample projectRemainingRegion(SearchSpace space) {
        CharacteristicSample sample = (CharacteristicSample)space.getTokenlist();
        return sample.getProjectedSampleRemainingAfterPos(dir, space.getPosReachedOnTokenlist());
    }
    
}
