/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * Labeller.java
 *
 * Created on 30 novembre 2003, 11.38
 * @author  Valter Crescenzi, Luigi Arlotta
 */
package roadrunner.labeller;

import java.util.*;
import java.util.logging.*;

import roadrunner.config.Config;
import roadrunner.config.Constants;
import roadrunner.WrappersRepository;
import roadrunner.ast.*;
import roadrunner.parser.BindingException;
import roadrunner.Wrapper;
/**
 *
 * @author   Luigi Arlotta, Valter Crescenzi
 */
public class Labeller extends VisitorPlugger implements Visitor, Constants {
    
    static private Logger log = Logger.getLogger(Labeller.class.getName());
    
    private WrappersRepository repository;
    private VisualLabeller visualLabeller;
    private PrefixLabeller prefixLabeller;
    private LabelExtractor labelExtractor;
    private boolean prefixEnabled;
    private boolean visualEnabled;
        
    /** Creates a new instance of Labeller */
    public Labeller(WrappersRepository repository) throws BindingException {
        this.repository = repository;
        this.prefixEnabled = Config.getPrefs().getBoolean(PREFIX_ENABLED);
        this.visualEnabled = Config.getPrefs().getBoolean(VISUAL_ENABLED);
        this.labelExtractor = new LabelExtractor();
        init();
    }
    
    private boolean isLabellingEnabled() {
        return this.prefixEnabled || this.visualEnabled;
    }    
    private boolean isLabellingByPrefixEnabled() {
        return this.prefixEnabled;
    }
    private boolean isVisualLabellingEnabled() {
        return this.visualEnabled;
    }

    public void init() throws BindingException {
        Iterator it = this.repository.getWrappers().iterator();
        while (it.hasNext()) {
            // first set anonymous labels for all variants...
            Wrapper wrapper = (Wrapper)it.next();
            wrapper.getExpression().getRoot().jjtAccept(
            new VisitorPlugger() {
                private int counter = 0;
                public boolean visit(ASTVariant variant) {
                    variant.setLabel(getAnonymousLabel(counter++));
                    return true;
                }
            }
            );
            // ...then set all labels it can guess
            Set instances = this.repository.getInstances(wrapper);
            if (isLabellingByPrefixEnabled()) {
                this.prefixLabeller = new PrefixLabeller(wrapper,instances);
            }
            if (isVisualLabellingEnabled()) {
                this.visualLabeller = new VisualLabeller(wrapper,instances);
            }
            if (isLabellingEnabled()) {
                wrapper.getExpression().getRoot().jjtAccept(this); //go labeller, go!
            }
        }
    }
    
    private String getAnonymousLabel(int attNumber) {
        final int range = ((int)'Z'-(int)'A')+1;
        int letter = attNumber % range;
        int number = attNumber / range;
        return "_"+((char)(letter+(int)'A'))+(number==0?"":Integer.toString(number))+"_";
    }
    
    public boolean visit(ASTVariant node) {
        String label = guessLabel(node);
        if (label!=null) node.setLabel(label);
        return true;
    }
        
    private String guessLabel(ASTVariant node) {
        // we can guess only label of text tokens
        if (!node.getToken().isPCDATA()) return null; 
        String label = null;
        //look for  constant texts whose occurrences are visually close to the occurrences of the variant values
        if (isVisualLabellingEnabled()) {
            label = this.visualLabeller.getLabel(node);
        }
        //look for a common suffix or prefix amongst all the values of a variants        
        if (label==null && isLabellingByPrefixEnabled()) {
            label = this.prefixLabeller.getLabel(node);
        }
        return this.labelExtractor.extractLabel(label);
    }
    
    
}
