/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * FreeTextHunter.java
 *
 * Created on
 * @author  Valter Crescenzi
 */
package roadrunner.parser;

import java.util.Set;
import java.util.HashSet;
import java.util.logging.*;

import org.w3c.dom.*;

import roadrunner.util.DOMLoader;
import roadrunner.config.Constants;
import roadrunner.config.Preferences;

import roadrunner.labeller.Box;

/** Hunts for free-text chunks
 * A free-text chunck includes a list of sibling sub-trees, at least
 * one of which is either a PCDATA or one of variant tags */

public class FreeTextHunter implements Constants {
    
    static private Logger log = Logger.getLogger(FreeTextHunter.class.getName());
    
    private Document document; //input DOM document
    private Document result;   //output DOM document
    private Preferences prefs; //preferences to config freetext collapsing
    private Set boxes;         //boxes detected during freetext collapsing
    
    private boolean keepTags;
    
    public FreeTextHunter(Document document, Preferences prefs) {
        this.document = document;
        this.prefs = prefs;
        this.keepTags = this.prefs.getBoolean(KEEP_TAGS);
    }
    
    public Document getCollapsedDocument() {
        String namespace = this.document.getNamespaceURI();     
        this.result = this.document.getImplementation().createDocument(namespace,null,null);
        build(this.document.getDocumentElement(), this.result);
        this.result.normalize();
        return this.result;
    }
    
    // Build another DOM with just one text node for each chunk of freetext
    void build(Node input, Node output) {
        if (input==null) return;
        
        if (isFreeTextChunk(input)) {
            // collapse sibling freetext chunks and eventually compute a containing bb
            StringBuffer freetext = new StringBuffer();
            Set bboxes = new HashSet();
            Node node = input;
            do  {
                collapseFreetext(node, freetext, bboxes);
                node = node.getNextSibling();
            } while (node!=null && isFreeTextChunk(node));
            Box containingBox = Box.getContainingBox(bboxes);
            if (containingBox!=null) {
                // encode bb as comment
                output.appendChild(this.result.createComment(Box.encode(containingBox)));
            }
            // freetext chunk in a new freetext node
            output.appendChild(this.result.createTextNode(freetext.toString()));
            // then visit next sibling
            build(node,output);
        }
        else {
            Node clone = this.result.importNode(input,false);
            output.appendChild(clone);
            // first visit childs
            build(input.getFirstChild(), clone);
            // then visit siblings
            build(input.getNextSibling(),output);
        }
    }
    
    private boolean isFreeTextChunk(Node node) {
        int type = node.getNodeType();
        switch (type) {
            case Node.TEXT_NODE:
                return true;
            case Node.COMMENT_NODE:
                return true;  // comments do not prevent from collapsing freetext
            case Node.ELEMENT_NODE:
                NodeList children = node.getChildNodes();
                if (children.getLength()==0) {
                    return isVariantTag(node); /* leaf tag */
                }
                else if (!isPresentationTag(node)) {
                    return false;
                }
                //it is one of freetext tags; consider it  a chunk
                //of freetext iff all children are freetext chunks
                else return allChildrenAreFreeTextChunks(node);
        }
        return false; //all other types of DOM node types stop freetext collapsing
    }
    
    private boolean allChildrenAreFreeTextChunks(Node node) {
        NodeList children = node.getChildNodes();
        int len = children.getLength();
        for (int i=0; i<len; i++) {
            if (!isFreeTextChunk(children.item(i)))
                return false;
        }
        return true;
    }
    
    /** One string for one subtree, recursively. */
    private void collapseFreetext(Node node, StringBuffer freetext, Set boxes) {
        int type = node.getNodeType();
        switch ( type ) {
            case Node.TEXT_NODE:
                freetext.append(node.getNodeValue());
                break;
            case Node.COMMENT_NODE:
                Comment comment = (Comment)node;
                if (Box.isBoundingBoxEncoding(comment.getNodeValue()))
                    boxes.add(Box.decode(comment.getNodeValue()));
                break;
            case Node.ELEMENT_NODE:
                String element = node.getNodeName();
                NodeList children = node.getChildNodes();
                int len = children.getLength();
                if (len>0) {
                    freetext.append(openTag(element));
                    for (int i=0; i<len; i++) {
                        collapseFreetext(children.item(i),freetext,boxes);
                    }
                    freetext.append(closeTag(element));
                }
                else freetext.append(leafTag(element));
                break;
        }
    }
    
    private String openTag(String element) {
        if (!keepTagsEnabled()) return " ";
        return '<'+element+'>';
    }
    private String closeTag(String element) {
        if (!keepTagsEnabled()) return " ";
        return "</"+element+'>';
    }
    private String leafTag(String element) {
        if (!keepTagsEnabled()) return " ";
        return '<'+element+"/>";
    }
    
    private boolean isPresentationTag(Node node) {
        int type = node.getNodeType();
        if (type!=Node.ELEMENT_NODE) return false;
        Set presTags = this.prefs.getSet(FREETEXT_TAGS);
        return presTags.contains(node.getNodeName().toLowerCase());
    }
    
    private boolean isVariantTag(Node node) {
        int type = node.getNodeType();
        if (type!=Node.ELEMENT_NODE) return false;
        Set varTags = this.prefs.getSet(VARIANT_TAGS);
        return varTags.contains(node.getNodeName().toLowerCase());
    }
    
    private boolean keepTagsEnabled() {
        return this.keepTags;
    }
    
}
